mod html_templte;

use std::{fs, io};
use std::fs::File;
use std::io::{BufRead, BufReader, Cursor, Read};
use sonic_channel::{Dest, IngestChannel, PushRequest, QueryRequest, SearchChannel, SonicChannel};
use xml::EventReader;
use std::collections::HashMap;
use std::path::Path;
use warp::Filter;
use std::io::Write;
use std::net::SocketAddr;
use xml::reader::XmlEvent;
use clap::{arg, command, value_parser};

async fn add_document(channel: &IngestChannel, collection: &str, bucket: &str, object: &str, text: &str) -> Result<(), Box<dyn std::error::Error>> {
    let dest = Dest::col_buc(collection, bucket).obj(object);
    let p = PushRequest::new(dest, text.to_string());
    channel.push(p)?;
    Ok(())
}

async fn search_document(server_address: String, server_password: String, collection: &str, bucket: &str, query: &str, limit: i32, offset: i32) -> Result<Vec<String>, Box<dyn std::error::Error>> {
    let search = SearchChannel::start(server_address, server_password)?;
    let results: Vec<String> = search.query(QueryRequest::new(
        Dest::col_buc(collection, bucket),
        query
    ).limit(limit as usize).offset(offset as usize)).unwrap_or_else(|_| { vec![] });
    search.quit()?;
    Ok(results)
}

async fn ingest_all(server_address: String, server_password: String, collection: String, ingest_file: String) -> std::io::Result<()>{
    let channel = IngestChannel::start(server_address, server_password)
        .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e))?;

    let ingested_names = retrieve_ingested_documents(ingest_file.clone()).unwrap();
    let binding = "./".to_owned() + &*ingest_file.clone();
    let path = Path::new(&binding);

    fs::remove_file(path).expect("Unable to delete");
    let mut file_ingested = File::create(path)?;
    for ingested_name in ingested_names.clone() {
        file_ingested.write_all(format!("{}\n", ingested_name).as_bytes()).expect(format!("Unable to write filename to {}", ingest_file).as_str());
    }

    let mut counter = 0;
    for entry in fs::read_dir("./")? {
        let entry = entry?;
        let list_name = entry.path();
        if list_name.is_dir() {
            let bucket = match list_name.file_name() {
                None => {
                    "unnamed list"
                },
                Some(file_name) => {
                    file_name.to_str().unwrap()
                }
            };

            for mail in fs::read_dir(list_name.clone().join("threads"))? {
                let mail = mail?;
                let path = mail.path();
                match path.extension() {
                    None => {}
                    Some(ext) => {
                        if ext.eq_ignore_ascii_case("xml") {
                            if ingested_names.contains(&path.file_stem().unwrap().to_str().unwrap().to_string()){
                                continue;
                            }
                            let file = File::open(path.clone());
                            let mut xml = String::new();
                            file.unwrap().read_to_string(&mut xml).expect("Unable to read a file");

                            let obj = path.file_stem().unwrap().to_str().unwrap();

                            println!("obj ingested({}): {}", counter, obj);
                            file_ingested.write_all(format!("{}\n", obj).as_bytes()).expect(format!("Unable to write filename to {}", ingest_file).as_str());
                            counter += 1;

                            let content = get_content_from_xml(&xml);
                            let contents = slice_string_into_chunks(&*content);

                            for c in contents {
                                add_document(&channel, &*collection, bucket, obj, c).await.expect("Unable to ingest email to Sonic Search server");
                            }
                        }
                    }
                }
            }
        }
    }
    channel.quit().map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e))?;

    Ok(())
}

fn slice_string_into_chunks(s: &str) -> Vec<&str> {
    let mut chunks = Vec::new();
    let mut start = 0;
    let len = s.len();

    while start < len {
        let end = if start + 15000 > len {
            len
        } else {
            let mut end = start + 15000;
            while !s.is_char_boundary(end) && end > start {
                end -= 1;
            }
            end
        };

        if start != end {
            chunks.push(&s[start..end]);
        }
        start = end;
    }

    chunks
}

fn get_content_from_xml(xml_data: &str) -> String {
    let cursor = Cursor::new(xml_data);
    let parser = EventReader::new(cursor);
    let mut in_content = false;
    let mut contents = Vec::new();
    let mut current_content = String::new();

    for e in parser {
        match e {
            Ok(XmlEvent::StartElement { name, .. }) if name.local_name == "content" => {
                in_content = true;
                current_content.clear(); // Clear any previous content
            },
            Ok(XmlEvent::Characters(chars)) if in_content => {
                let cleaned_chars = chars.chars()
                    .filter(|c| c.is_alphabetic() || *c == ' ')
                    .collect::<String>();
                current_content.push_str(&cleaned_chars);
            },
            Ok(XmlEvent::EndElement { name }) if name.local_name == "content" => {
                if in_content {
                    contents.push(current_content.clone());
                    in_content = false;
                }
            },
            Err(_) => {
                break;
            },
            _ => {}
        }
    }

    contents.join("")
}

fn get_lists() -> Result<Vec<String>, io::Error>{
    let mut result: Vec<String> = vec![];

    match fs::read_dir("./") {
        Ok(entries) => {
            for entry in entries {
                match entry {
                    Ok(entry) => {
                        let list_name = entry.path();
                        if list_name.is_dir() {
                            if let Some(name) = list_name.file_name().expect("Unable to get file name of the entry of the lists").to_str() {
                                result.push(name.to_string());
                            }
                        }
                    }
                    Err(_err) => {
                        return Err(_err);
                    }
                }
            }
        }
        Err(_err) => {
            return Err(_err);
        }
    }

    Ok(result)
}

async fn get_lists_warp() -> Result<impl warp::Reply, warp::Rejection>{
    match get_lists() {
        Ok(result) => Ok(warp::reply::json(&result)),
        Err(_err) => Err(warp::reject::not_found())
    }
}

async fn handle_search(query: HashMap<String, String>, (server_address, server_password): (String, String)) -> Result<impl warp::Reply, warp::Rejection> {
    if let Some(q) = query.get("q") {
        let list = match query.get("list"){
            None => {""}
            Some(list) => {list}
        };
        let limit = match query.get("limit") {
            None => 10,
            Some(limit) => limit.parse().unwrap_or_else(|_| 10)
        };
        let offset = match query.get("offset") {
            None => 0,
            Some(offset) => offset.parse().unwrap_or_else(|_| 0)
        };
        let results = search_document(server_address, server_password, "mailbox", &*list, q, limit, offset).await.unwrap();
        Ok(warp::reply::json(&results))
    } else {
        Err(warp::reject::not_found())
    }
}

fn retrieve_ingested_documents(ingest_file: String) -> io::Result<Vec<String>> {
    let binding = "./".to_owned() + &*ingest_file;
    let path = Path::new(&binding);

    if !path.exists() {
        File::create(&path)?;
    }

    let file = File::open(&path)?;
    let reader = BufReader::new(file);

    let lines: Vec<String> = reader.lines()
        .filter_map(|line| line.ok())
        .collect();

    Ok(lines)
}

async fn run_server(server_address: String, server_password: String, collection: String, ingest_file: String, gui_address: SocketAddr){
    let server_address_clone = server_address.clone();
    let server_password_clone = server_password.clone();
    let collection_clone = collection.clone();
    let ingest_file_clone = ingest_file.clone();

    let index = warp::get()
        .and(warp::path::end())
        .map(move || {
            warp::reply::html(html_templte::HTML_PAGE)
        });

    let search_route = warp::get()
        .and(warp::path("search"))
        .and(warp::query::<HashMap<String, String>>())
        .and(warp::any().map(move || (server_address_clone.clone(), server_password_clone.clone())))
        .and_then(handle_search);

    let get_lists_route = warp::get()
        .and(warp::path("get-lists"))
        .and_then(get_lists_warp);

    let routes = index.or(search_route).or(get_lists_route);

    ingest_all(server_address, server_password, collection_clone.clone(), ingest_file_clone.clone()).await.expect("Unable to ingest all emails");

    println!("HTTP server is running on http://{}...", gui_address);
    warp::serve(routes).run(gui_address).await;
    println!("HTTP server closed");
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let default_server_address = "127.0.0.1:5041";
    let default_server_password = "SecretPassword";
    let default_collection = "mailbox";
    let default_ingest_file = ".sonicSearchIngested.txt";
    let default_gui_address = "127.0.0.1:3030";

    let matches = command!()
        .arg(arg!([name] "Optional name to operate on"))
        .arg(arg!(-a --address <STRING> "Sets the server address and port")
            .required(false)
            .value_parser(value_parser!(String))
            .default_value(default_server_address)
            .default_missing_value(default_server_address)
        )
        .arg(arg!(-p --password <STRING> "Sets the server password")
            .required(false)
            .value_parser(value_parser!(String))
            .default_value(default_server_password)
            .default_missing_value(default_server_password)
        )
        .arg(arg!(-c --collection <STRING> "Sets the collection for all emails")
            .required(false)
            .value_parser(value_parser!(String))
            .default_value(default_collection)
            .default_missing_value(default_collection)
        )
        .arg(arg!(-f --file <STRING> "Sets the file to store names of ingested files")
            .required(false)
            .value_parser(value_parser!(String))
            .default_value(default_ingest_file)
            .default_missing_value(default_ingest_file)
        )
        .arg(arg!(-g --gui <STRING> "Sets the address and port for gui to operate on")
            .required(false)
            .value_parser(value_parser!(String))
            .default_value(default_gui_address)
            .default_missing_value(default_gui_address)
        )
        .get_matches();

    let server_address = match matches.get_one::<String>("address") {
        None => default_server_address,
        Some(address) => address
    };

    let server_password = match matches.get_one::<String>("password") {
        None => default_server_password,
        Some(password) => password
    };

    let collection = match matches.get_one::<String>("collection") {
        None => default_collection,
        Some(collection) => collection
    };

    let ingest_file = match matches.get_one::<String>("file") {
        None => default_ingest_file,
        Some(file) => file
    };

    let gui_address: SocketAddr = match matches.get_one::<String>("gui") {
        None => default_gui_address.parse().expect("Unable to parse default address for gui"),
        Some(address) => address.parse().expect("Unable to parse address for gui")
    };

    run_server(
        server_address.to_string().clone(),
        server_password.to_string().clone(),
        collection.to_string().clone(),
        ingest_file.to_string().clone(),
        gui_address.clone()
    ).await;
    Ok(())
}